/* @(#)indexed.pl	1.7 29 Sep 1993

Here we supply the basic database operations for indexed code. This file
is meant to be used with code that uses the term expansions defined in
either dynamic_indexer.pl or module_indexer.pl.

See the README in this directory.

This software is free: please copy it and/or modify it. Do with it what
you like.  It is a gift from KnowledgeWare Inc, and was written by Tom
Howland.

*/

:- module(indexed, [
	indexed_assert/1,
	indexed_retract/1,
	indexed_retractall/1,
	indexed_module_initialize/1	% described below
   ]).

:- meta_predicate
	indexed_assert(0),
	indexed_retract(0),
	indexed_retractall(0),
	indexed_module_initialize(0).

:- use_module(messages(language('QU_messages'))).
:- use_module(library(types), [must_be/4]).

:- multifile 'QU_messages':operation/3.
:- multifile 'QU_messages':typename/3.

indexed_assert(RawHead) :-
    check_module_prefixes(RawHead, user, Mod, Head),
    must_be(callable, Head, 1, indexed_assert(RawHead)),
    (   Mod:'** indexed **'(Head)
    ->  Mod:'** indexed assert **'(Head)
    ;   non_indexed_pred(Mod, Head, indexed_assert, indexed_assert(RawHead))
    ).

indexed_retract(RawHead) :-
    check_module_prefixes(RawHead, user, Mod, Head),
    must_be(callable, Head, 1, indexed_retract(RawHead)),
    (   Mod:'** indexed **'(Head)
    ->  Mod:'** indexed retract **'(Head)
    ;   non_indexed_pred(Mod, Head, indexed_retract, indexed_retract(RawHead))
    ).

indexed_retractall(RawHead) :- indexed_retract(RawHead), fail ; true.

check_module_prefixes(ModTerm, _, _, _) :-
        var(ModTerm),
        !,
        raise_exception(instantiation_error(ModTerm, 0)).
check_module_prefixes(Mod:MTerm, _, Module, Term) :-
        !,
        check_module_prefixes(MTerm, Mod, Module, Term).
check_module_prefixes(Term, Mod, Mod, Term).
              
non_indexed_pred(Mod, Head, Goal, Action) :-
    functor(Head, Name, Arity),
    raise_exception(permission_error(Goal, Action, non_indexed_pred,
				     Mod:Name/Arity, 0)).

'QU_messages':typename(non_indexed_pred) --> ['non-indexed predicate'-[]].

'QU_messages':operation(indexed_assert) --> ['indexed assertion'-[]].
'QU_messages':operation(indexed_retract) --> ['indexed retraction'-[]].

% This is the only way I can think of to get module-partioned facts
% declared dynamic.

% indexed_module_initialize(+RawHead) should be passed a new module
% name. It then sweeps through the facts that may be called in that
% module and ensures that they are declared dynamic.

indexed_module_initialize(RawHead) :-
    check_module_prefixes(RawHead, user, Mod, Head),
    Mod:'** module initialization **'(N, A),
    functor(F, N, A),
    assert(Head:F, Ref), erase(Ref),
    fail ; true.
