/* @(#)indexer.pl	1.8 16 Mar 1994

Since this file is only used by term_expansion/2, it should be loaded with

:- load_files(indexer(indexer), when(compile_time),if(changed)]).

The ":- index" directive supplied by this file is described in the
README file in this directory.

This module is intended for use in indexing compiled tables of prolog
facts. If you wish to index dynamic facts, see dynamic_indexer.pl.

This software is free: please copy it and/or modify it. Do with it what
you like.  It is a gift from KnowledgeWare Inc, and was written by Tom
Howland.

*/

:- module(indexer, []).
:- use_module(low).
:- use_module(library(addportray)).
:- dynamic indexed/4.

indexer_term_expansion(Rule, [NewRule|Clauses]) :-
    get_prolog_load_module(Mod),
    head(Rule, Head),
    functor(Head, Name, Arity),
    indexed(Head, Mod, ClNum, NewHead),
    retract(clause_count(Name, Arity, Mod, ClNum)),
    X is ClNum+1,
    assert(clause_count(Name, Arity, Mod, X)),
    new_rule(Rule, NewHead, NewRule),
    findall(Ming, '** hash index **'(Head, Mod, ClNum, Ming), Clauses).
indexer_term_expansion((:-index(Specs)), Clauses) :-
    index_clauses(Specs, _, [], Head, NewHead, Mod, NewName,
		  NewArity, FirstArg, Body, Hash, Hash),
    functor(Head, Name, Arity),
    assert(clause_count(Name, Arity, Mod, 0)),
    assert(indexed(Head, Mod, FirstArg, NewHead)),
    parse_index_specs(Specs, Head, Mod, Mod, FirstArg, Body, _, _, _, Dis,
		      Com, _, _, _, _),
    append(Dis, [(:-discontiguous NewName/NewArity), (Head:-Body,NewHead)],
	   Clauses),
    assert_each(Com).

new_rule((_:-Y), N, (N:-Y)) :- !.
new_rule(_, N, N).

get_prolog_load_module(Mod) :-
    (   prolog_load_context(module, Mod)
    ->  true
    ;   Mod = user
    ).

assert_each([]).
assert_each([H|T]) :- assert(H), assert_each(T).

:- initialization
    add_linking_clause(indexer_term_expansion, term_expansion, 2),
    op(1150, fx, index).

end_of_file.

t(X) :- expand_index_directive((:-index(foo(*))), X).

dbg(X) :-
    prt(call, X),
    (   X,
	(   prt(exit, X)
	;   prt(redo, X), fail
	)
    ;   prt(fail, X), fail
    ).
    
prt(A, X) :- format('~N~a ~q~n', [A, X]).
